package wpd

import (
	"bytes"
	"context"
	"errors"
	"fmt"
	"io/ioutil"

	"github.com/rs/zerolog"
	"orus.io/orus-io/rednerd/rendering"
)

// BufferCloser is a bytes.Buffer that implements io.ReadCloser
type BufferCloser struct {
	*bytes.Buffer
}

// Close does nothing
func (b BufferCloser) Close() error {
	return nil
}

// Engine implements rendering.Engine
type Engine struct {
	wpd *WPD
}

// Info returns the engine description
func (e Engine) Info() rendering.EngineInfo {
	return rendering.EngineInfo{
		Type:        "Weasyprint",
		Description: `Render HTML to PDF with weasyprint.`,
		Transitions: []rendering.EngineTransitionInfo{
			{FromType: "text/html", ToType: "application/pdf", Priority: 2.0},
		},
		Metadata: []rendering.EngineMetadataInfo{
			{
				Name:        "weasyprint-css",
				Description: `A global CSS passed to weasyprint `,
			}, {
				Name: "print-paper-size",
				Description: `A page size ("A4", "A5 landscape", "21cm 29.6cm"...).
Will be converted to CSS and combined to weasyprint-css if provided`,
			}, {
				Name: "print-paper-margin",
				Description: `
A page margin ("1cm", "1cm 5cm", "1cm 2cm 1cm 2cm", "top right bottom left"...).
Will be converted to CSS and combined to weasyprint-css if provided
			`,
			},
		},
	}
}

// Render a document from a html doc to pdf
func (e Engine) Render(
	ctx context.Context, w rendering.DocumentWriter, r rendering.DocumentReader, toType string,
) error {
	if toType != "application/pdf" {
		return errors.New("invalid output type: " + toType)
	}

	mimetype, metadata, err := r.Metadata()
	if err != nil {
		return err
	}
	if mimetype != "text/html" {
		return errors.New("invalid input type: " + mimetype)
	}

	docname := metadata["title"]
	css := metadata["weasyprint-css"]

	if pageSize := metadata["print-paper-size"]; pageSize != "" {
		css += fmt.Sprintf(`
@page {
	size: %s;
}
`,
			pageSize)
	}

	if pageMargin := metadata["print-paper-margin"]; pageMargin != "" {
		css += fmt.Sprintf(`
@page {
	margin: %s;
}
`,
			pageMargin)
	}

	html, err := ioutil.ReadAll(r)
	if err != nil {
		return err
	}

	zerolog.Ctx(ctx).Trace().Str("css", css).Str("html", string(html)).Msg("will render")
	output, err := e.wpd.Render(ctx, docname, css, html)
	if err != nil {
		return err
	}
	w.Metadata().Update(metadata)
	if _, err := w.Write(output); err != nil {
		return err
	}
	return nil
}

// NewEngine creates a wpd rendering engine
func NewEngine(log zerolog.Logger) (*Engine, error) {
	wpd, err := NewWPD(RealRunnerFactory, log)
	if err != nil {
		return nil, err
	}
	return &Engine{
		wpd: wpd,
	}, nil
}
