package wpd

import (
	"bufio"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"sync"
	"sync/atomic"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
)

var (
	ErrWpdCrashed   = errors.New("wpd crashed")
	ErrShuttingDown = errors.New("wpd is shutting down")
)

// Runner ...
type Runner interface {
	Stdin() io.WriteCloser
	Stdout() io.ReadCloser
	Stop() error
}

// RunnerInfo ...
type RunnerInfo struct {
	Wpd      string `json:"wpd"`
	Version  string `json:"version"`
	Protocol string `json:"protocol"`
}

// RunnerFactory is a function that creates a Runner
type RunnerFactory func() (Runner, error)

// Job ...
type Job struct {
	Docname string       `json:"-"`
	ID      uint64       `json:"id"`
	CSS     string       `json:"css"`
	HTML    []byte       `json:"-"`
	Size    int          `json:"size"`
	Result  chan *Result `json:"-"`
	started time.Time
	context context.Context
	stop    chan struct{}
}

func (job *Job) success(r *Result) {
	close(job.stop)
	renderDocumentDuration.WithLabelValues(
		job.Docname, "success",
	).Observe(float64(time.Since(job.started)) / float64(time.Millisecond))
	job.Result <- r
}

func (job *Job) failure(err error) {
	close(job.stop)
	renderDocumentDuration.WithLabelValues(
		job.Docname, "success",
	).Observe(float64(time.Since(job.started)) / float64(time.Millisecond))
	job.Result <- &Result{ID: job.ID, Error: err}
}

// Result ...
type Result struct {
	ID      uint64 `json:"id"`
	Content []byte `json:"-"`
	Size    int    `json:"size"`
	Error   error  `json:"error"`
}

// WPD ...
type WPD struct {
	wg      sync.WaitGroup
	queue   chan *Job
	results chan *Result

	factory RunnerFactory
	runner  Runner

	counter uint64

	quit chan struct{}

	log zerolog.Logger
}

// WriteAll writes a list of []byte to a writer in one shot
func WriteAll(w io.Writer, data ...[]byte) (int, error) {
	var n int
	for _, d := range data {
		nn, err := w.Write(d)
		n += nn
		if err != nil {
			return n, err
		}
	}
	return n, nil
}

func heloRunner(runner Runner, reader *bufio.Reader, log zerolog.Logger) (err error) {
	t := time.AfterFunc(time.Second*5, func() {
		if err := runner.Stop(); err != nil {
			log.Err(err).Msg("Error stopping the runner")
		}
	})
	defer func() {
		if !t.Stop() {
			err = fmt.Errorf("timeout reading runner info header")
		}
	}()

	line, err := reader.ReadBytes(byte('\n'))
	if err != nil {
		fmt.Println(err)
		return err
	}
	var info RunnerInfo
	if err := json.Unmarshal(line, &info); err != nil {
		panic(err)
	}
	if info.Wpd == "" {
		return fmt.Errorf("invalid wpd info: missing \"wpd\"")
	}
	if info.Protocol != "1" {
		return fmt.Errorf("invalid wpd protocol version. Expects \"1\", got: \"%s\"", info.Protocol)
	}
	return nil
}

// NewWPD ...
func NewWPD(runnerFactory RunnerFactory, log zerolog.Logger) (*WPD, error) {
	var wpd = WPD{
		queue:   make(chan *Job),
		results: make(chan *Result),
		factory: runnerFactory,
		quit:    make(chan struct{}),
		log:     log,
	}

	runner, reader, err := wpd.lauchRunner()

	if err != nil {
		return nil, err
	}

	wpd.wg.Add(1)
	go wpd.main(runner, reader)

	return &wpd, nil
}

func (wpd *WPD) lauchRunner() (Runner, *bufio.Reader, error) {
	runner, err := wpd.factory()
	if err != nil {
		return nil, nil, err
	}
	reader := bufio.NewReader(runner.Stdout())
	if err := heloRunner(runner, reader, wpd.log); err != nil {
		if err := runner.Stop(); err != nil {
			wpd.log.Err(err).Msg("Error stopping the runner")
		}
		return nil, nil, err
	}
	return runner, reader, nil
}

// Render ...
func (wpd *WPD) Render(ctx context.Context, docname, css string, html []byte) ([]byte, error) {
	job := wpd.Submit(ctx, docname, css, html)
	result := <-job.Result
	return result.Content, result.Error
}

// Submit ...
func (wpd *WPD) Submit(ctx context.Context, docname, css string, html []byte) *Job {
	id := atomic.AddUint64(&wpd.counter, 1)

	var job = Job{
		Docname: docname,
		ID:      id,
		CSS:     css,
		HTML:    html,
		Size:    len(html),
		Result:  make(chan *Result, 1),
		started: time.Now(),
		context: ctx,
		stop:    make(chan struct{}),
	}
	wpd.queue <- &job
	return &job
}

// Close ..
func (wpd *WPD) Close() error {
	close(wpd.quit)
	// now close the runner
	runnerStopErr := wpd.runner.Stop()
	wpd.wg.Wait()
	close(wpd.queue)
	close(wpd.results)
	for range wpd.queue {
	}
	for range wpd.results {
	}
	return runnerStopErr
}

func (wpd *WPD) main(runner Runner, reader *bufio.Reader) {
	defer wpd.wg.Done()

	for {
		running, err := wpd.run(runner, reader)
		if err == nil {
			return
		}
		if errors.Is(err, ErrShuttingDown) {
			log.Info().Msg("engine is shutting down, signal jobs and exit")
			for _, job := range running {
				job.failure(err)
			}
			return
		}
		wpd.log.Err(err).Int("running jobs", len(running)).Msg("run failed")

		runner, reader, err = wpd.lauchRunner()
		if err != nil {
			wpd.log.Err(err).Msg("lauchRunner failed")
			for _, job := range running {
				job.failure(ErrWpdCrashed)
			}
			return
		}
	}
}

func (wpd *WPD) run(runner Runner, reader *bufio.Reader) ([]*Job, error) {
	defer runner.Stdin().Close()

	go wpd.readWpdOutput(reader)

	var (
		running      = make(map[uint64]*Job)
		canceledJobs = make(chan *Job)
	)

	runningList := func() []*Job {
		var runningList []*Job
		for _, j := range running {
			runningList = append(runningList, j)
		}
		return runningList
	}

	for {
		select {
		case <-wpd.quit:
			return runningList(), ErrShuttingDown
		case job := <-wpd.queue:
			if b, err := json.Marshal(job); err != nil {
				job.failure(err)
			} else {
				if _, err := WriteAll(
					wpd.runner.Stdin(), b, []byte("\n"), job.HTML, []byte("\n"),
				); err != nil {
					// TODO handle write errors
					job.failure(err)
				} else {
					running[job.ID] = job
					go func() {
						select {
						case <-wpd.quit:
							return
						case <-job.stop:
							return
						case <-job.context.Done():
							canceledJobs <- job
						}
					}()
				}
			}
		case result := <-wpd.results:
			if job, ok := running[result.ID]; ok {
				job.success(result)
				delete(running, result.ID)
			}
		case job := <-canceledJobs:
			if job, ok := running[job.ID]; ok {
				delete(running, job.ID)
				job.failure(job.context.Err())

				b, err := json.Marshal(map[string]interface{}{
					"id":     job.ID,
					"action": "cancel",
				})
				if err != nil {
					panic(err)
				}
				if _, err := WriteAll(
					wpd.runner.Stdin(), b, []byte("\n"),
				); err != nil {
					// TODO handle write errors
					wpd.log.Err(err).Msg("could not write to wpd")
				}
			}
		}
	}
}

func (wpd *WPD) readWpdOutput(reader *bufio.Reader) {
	defer wpd.wg.Done()

	for {
		line, err := reader.ReadBytes(byte('\n'))
		if err != nil {
			wpd.log.Err(err).Msg("could not read bytes from wpd stdout")
			return
		}
		var r Result
		if err := json.Unmarshal(line, &r); err != nil {
			b := line
			if len(b) > 1024 {
				b = b[:1024]
			}
			wpd.log.Err(err).Str("result", string(b)).Msg("could not unmarshal wpd result")
			// TODO emergency exit with error, maybe restart the runner
			panic(err)
		}
		if r.Size != 0 {
			var buf = make([]byte, r.Size)
			_, err := io.ReadFull(reader, buf)
			if err != nil {
				wpd.log.Err(err).Msg("could not ready wpd result body")
				// TODO emergency exit with error, maybe restart the runner
				panic(err)
			}

			if b, err := reader.ReadByte(); err != nil || b != byte('\n') {
				wpd.log.Err(err).Msg("could not ready wpd result body last byte")
				// TODO emergency exit with error, maybe restart the runner
				panic(err)
			}
			r.Content = buf
		}
		wpd.results <- &r
	}
}

var (
	renderDocumentDuration = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "alpe_wpd_render",
			Help:       "The WPD rendering duration and count",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
		},
		[]string{"docname", "status"},
	)
)
