package rendering

import (
	"context"
	"errors"

	"orus.io/orus-io/rednerd/models"
)

// ErrUnknownTemplateLanguage is returned if no engine supports the given language
var ErrUnknownTemplateLanguage = errors.New("unknown template language")

// TemplateEngineInfo describes an engine capabilities
type TemplateEngineInfo struct {
	Type        string
	Language    string
	Description string
	Metadata    []EngineMetadataInfo
}

// Template is a parsed and loaded template, ready to be rendered
type Template interface {
	Varlist() ([]string, error)
	Render(ctx context.Context, w DocumentWriter, record models.Record) error
}

// TemplateEngine is able to render a document
type TemplateEngine interface {
	Info() TemplateEngineInfo
	Load(ctx context.Context, template *models.Template) (Template, error)
}

// RegisterTemplateEngine a new engine factory
func RegisterTemplateEngine(name string, engine TemplateEngine) error {
	return DefaultTemplateEngineRegistry.Register(name, engine)
}

// MustRegisterTemplateEngine registers an TemplateEngine and panics if it fails
func MustRegisterTemplateEngine(name string, engine TemplateEngine) {
	DefaultTemplateEngineRegistry.MustRegister(name, engine)
}

// TemplateEngineRegistry is a set of TemplateEngine, ready for use
type TemplateEngineRegistry struct {
	TemplateEngines map[string]TemplateEngine
}

// Register a TemplateEngine
func (r *TemplateEngineRegistry) Register(name string, engine TemplateEngine) error {
	if _, ok := r.TemplateEngines[name]; ok {
		return ErrDuplicateName
	}
	r.TemplateEngines[name] = engine
	return nil
}

// MustRegister a TemplateEngine
func (r *TemplateEngineRegistry) MustRegister(name string, engine TemplateEngine) {
	if err := r.Register(name, engine); err != nil {
		panic(err)
	}
}

// Find finds a template engine by the language it supports
func (r *TemplateEngineRegistry) Find(language string) TemplateEngine {
	for _, engine := range r.TemplateEngines {
		if engine.Info().Language == language {
			return engine
		}
	}
	return nil
}

// Load a template with the proper engine
func (r *TemplateEngineRegistry) Load(ctx context.Context, template *models.Template) (Template, error) {
	engine := r.Find(template.Language)
	if engine == nil {
		return nil, ErrUnknownTemplateLanguage
	}
	return engine.Load(ctx, template)
}

// NewTemplateEngineRegistry creates a TemplateEngineRegistry
func NewTemplateEngineRegistry() *TemplateEngineRegistry {
	return &TemplateEngineRegistry{
		TemplateEngines: make(map[string]TemplateEngine),
	}
}

var (
	// DefaultTemplateEngineRegistry is the default engine registry
	DefaultTemplateEngineRegistry = NewTemplateEngineRegistry()
)

func (r *TemplateEngineRegistry) NewPipeline(
	ctx context.Context,
	registry *EngineRegistry, template *models.Template, metadata models.Metadata, targetType string,
) (*TemplatePipeline, error) {
	tmpl, err := r.Load(ctx, template)
	if err != nil {
		return nil, err
	}
	p, err := registry.GetPipeline(template.Produces, targetType)
	if err != nil {
		return nil, err
	}
	return &TemplatePipeline{
		templateProduces: template.Produces,
		metadata:         metadata,
		template:         tmpl,
		pipeline:         p,
	}, nil
}

type TemplatePipeline struct {
	templateProduces string
	metadata         models.Metadata
	template         Template
	pipeline         Pipeline
}

func (t *TemplatePipeline) Render(
	ctx context.Context, w DocumentWriter, record models.Record,
) error {
	ctx, cancel := context.WithCancel(ctx)
	defer cancel()
	pipeErr := make(chan error, 1)
	if len(t.pipeline) != 0 {
		pw, pr := DocumentPipe()
		go func(ctx context.Context, w DocumentWriter, r DocumentReader) {
			pipeErr <- t.pipeline.Render(ctx, w, r)
		}(ctx, w, pr)
		w = pw
	}
	w.SetType(t.templateProduces)
	if err := t.template.Render(ctx, w, record); err != nil {
		_ = w.Close()
		return err
	}
	if err := w.Close(); err != nil {
		return err
	}
	if len(t.pipeline) != 0 {
		return <-pipeErr
	}
	return nil
}
